@@ -1,24 +1,21 @@
+use anyhow::Error;
 use metagoblin::archive;
-
-use crate::Opt;
-
-use prettytable::Cell;
-use prettytable::Row;
+use prettytable::{row, Cell, Row};
 
 use crate::format::*;
+use crate::Opt;
 
-use std::io;
-
-pub struct Archive<'a> {
-    archive: archive::Archive<'a>,
+pub struct Archive<'archive> {
+    archive: archive::Archive<'archive>,
     args: Opt,
 }
 
-impl<'a> Archive<'a> {
-    pub fn new(archive: archive::Archive<'a>, args: Opt) -> Self {
+impl<'archive> Archive<'archive> {
+    pub fn new(archive: archive::Archive<'archive>, args: Opt) -> Self {
         Archive { archive, args }
     }
-    pub fn print(&self) -> io::Result<()> {
+
+    pub fn print(&self) -> Result<(), Error> {
         let archive = &self.archive;
         let args = &self.args;
         let color = args.color;
@@ -27,8 +24,8 @@ impl<'a> Archive<'a> {
         let mut symbol_table = new_table(row![br->"Symbol", rb->"Owner"]);
         for (membername, member, symbols) in archive.summarize() {
             table.add_row(Row::new(vec![
-                sz_cell(member.size() as u64),
-                Cell::new(&symbols.len().to_string()),
+                sz_cell(u64::try_from(member.size())?),
+                Cell::new(&format!("{}", symbols.len())),
                 str_cell(membername).style_spec("brFr"),
             ]));
 
@@ -42,6 +39,6 @@ impl<'a> Archive<'a> {
 
         print_table_to_stdout(&table, color)?;
         println!();
-        print_table_to_stdout(&symbol_table, color)
+        print_table_to_stdout(&symbol_table, color).map_err(Into::into)
     }
 }
